#    MIDAS Access Server  for  CAMAC 

global env
source [file join $env(MIDASBASE) TclHttpd tcl MIDASDomainHandler.tcl]       ;#  load default support procedures

#     Note that the method of CAMAC access (e.g. via the Hytec PCI 5992 interface) is defined in the server startup file

global MIDAS_CAMACAccessServer_Verbose
set MIDAS_CAMACAccessServer_Verbose 0

global MIDAS_RegisterAccessServer_Verbose
set MIDAS_RegisterAccessServer_Verbose 0


# Use namespaces to isolate your methods
namespace eval urn:CAMACAccessServer {

global env
source [file join $env(MIDASBASE) TclHttpd tcl RegisterAccessServer.tcl]       ;#  load common procedures


#    set up the Instrinsic registers

    set HasBeenConfigured 0
    set Address ""
    set Data 0
    set Status 0
    set Debug 0

    set RegisterIndex 1
    
    set RegisterNames(HasBeenConfigured) $RegisterIndex
    set R[set RegisterIndex](class) camState
    incr RegisterIndex

    set RegisterNames(Camac.Address) $RegisterIndex
    set R[set RegisterIndex](class) camAddr
    incr RegisterIndex

    set RegisterNames(Camac.Execute) $RegisterIndex
    set R[set RegisterIndex](class) camExec
    incr RegisterIndex

    set RegisterNames(Camac.Data) $RegisterIndex
    set R[set RegisterIndex](class) camData
    incr RegisterIndex

    set RegisterNames(Camac.Status) $RegisterIndex
    set R[set RegisterIndex](class) camStatus
    incr RegisterIndex

    set RegisterNames(Camac.Debug) $RegisterIndex
    set R[set RegisterIndex](class) camDebug
    incr RegisterIndex


#    xCAMAC register attributes
#    cnaf  -  normal CAMAC codes
#    q   
#    width (-w) = 16 16 bit data   (default)
#                 24 24 bit data
#    access     = ro
#               = wo
#               = rw         (default)
#    format     = x (hex)    (default)
#               = d (decimal)
#               = b (binary)

   set xCAMAC_options {
         {c.arg 1    "CAMAC crate"}
         {n.arg 1    "CAMAC station"}
         {a.arg 0    "CAMAC subaddress"}
         {f.arg 0    "CAMAC function"}
         {q.arg 0    "QX flag"}
         {w.arg 16   "data width"}
         {p.arg "rw" "access permission"}
         {b.arg 0    "field width (bits)"}
         {l.arg 0    "field length (bits)"}
         {z.arg "x"  "data format returned"}
         {i.arg 0xa55a  "initial value"}
   }

   set xCAMAC_Updateoptions {
         {c.arg 0xa55a    "CAMAC crate"}
         {n.arg 0xa55a    "CAMAC station"}
         {a.arg 0xa55a    "CAMAC subaddress"}
         {f.arg 0xa55a    "CAMAC function"}
         {q.arg 0xa55a    "QX flag"}
         {w.arg 0xa55a    "data width"}
         {p.arg 0xa55a    "access permission"}
         {b.arg 0xa55a    "field width (bits)"}
         {l.arg 0xa55a    "field length (bits)"}
         {z.arg 0xa55a    "data format returned"}
         {i.arg 0xa55a    "initial value"}
    }

    set xCAMAC_Usage ": xCAMAC  -c ? -n ? -a ? -f ? -q 0|1 -w 16|24 -p ro|wo|rw -b ? -l ? -z w|b|d -i ?"

#    cCAMAC register attributes
#    cnaf  -  normal CAMAC codes
#    q   


    set cCAMAC_options {
          {c.arg 1  "CAMAC crate"}
          {n.arg 1  "CAMAC station"}
          {a.arg 0  "CAMAC subaddress"}
          {f.arg 0  "CAMAC function"}
          {q.arg 0  "QX flag"}
    }

    set cCAMAC_Updateoptions {
          {c.arg 0xa55a  "CAMAC crate"}
          {n.arg 0xa55a  "CAMAC station"}
          {a.arg 0xa55a  "CAMAC subaddress"}
          {f.arg 0xa55a  "CAMAC function"}
          {q.arg 0xa55a  "QX flag"}
    }

    set cCAMAC_Usage ": cCAMAC  -c ? -n ? -a ? -f ? -q 0|1"

    proc AccessServerVerbose {option} {
        global MIDAS_CAMACAccessServer_Verbose
        global MIDAS_RegisterAccessServer_Verbose
        global MIDAS_SOAPDomainHandler_Verbose
        set MIDAS_CAMACAccessServer_Verbose $option
        set MIDAS_RegisterAccessServer_Verbose $option
        set MIDAS_SOAPDomainHandler_Verbose $option
        return [rpcvar::rpcvar string "0 OK"]
    }

    proc DefineRegister {Register Class Attributes} {
        global MIDAS_CAMACAccessServer_Verbose
        upvar RegisterIndex RegisterIndex
        upvar RegisterNames RegisterNames

        if {$MIDAS_CAMACAccessServer_Verbose} \
             {puts "called method CAMACAccessServer::DefineRegister with parameters $Register $Class $Attributes"}
             
#    Check Register name is not already in use

        if {[info exists RegisterNames($Register)]} then {
             return [rpcvar::rpcvar string "0x10006 Register name already in use"]
        }

#    Check Class is valid

        switch $Class {
            xCAMAC -
            cCAMAC  {}

            default {
               return [rpcvar::rpcvar string "0x10007 Class name is invalid or not supported"]
            }
        }

#    create register

        switch $Class {
            xCAMAC  {

                set RegisterNames($Register) $RegisterIndex

                upvar R$RegisterIndex R
                set R(class) $Class

                upvar xCAMAC_options options
                foreach item $options {
                    set opt [first [split [first $item] .]]
                    set R($opt) [second $item]
                }

                set R(LastRead)  0x0
                set R(LastWrite) 0x0

                incr RegisterIndex
            }
            cCAMAC  {

                set RegisterNames($Register) $RegisterIndex

                upvar R$RegisterIndex R
                set R(class) $Class

                upvar cCAMAC_options options

                foreach item $options {
                    set opt [first [split [first $item] .]]
                    set R($opt) [second $item]
                }

                incr RegisterIndex
            }
            default {
                return [rpcvar::rpcvar string "0x10004 Invalid request"]
            }
        }

#    apply attributes

        set response [do_WriteRegisterAttributes $Register $Attributes]

        return [rpcvar::rpcvar string $response]
    }


    proc do_ReadRegisterAttributes {Register} {

        global MIDAS_CAMACAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames

        set Index $RegisterNames($Register)
        upvar 2 R$Index R
        switch $R(class) {
            xCAMAC {
                 set attributes "-c$R(c) -n$R(n) -a$R(a) -f$R(f) -q$R(q) -w$R(w) -p$R(p) -l$R(l) -b$R(b) -z$R(z)"
                 if {$R(i) != 0xa55a} {append attributes " -i$R(i)"}
            }
            cCAMAC {
                 set attributes "-c$R(c) -n$R(n) -a$R(a) -f$R(f) -q$R(q)"
            }
            default {
                return "0x10004 Invalid request"
            }
        }

        return "0 $attributes"
    }


    proc do_WriteRegisterAttributes {Register Attributes} {

        global MIDAS_CAMACAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames

        set Index $RegisterNames($Register)
        upvar 2 R$Index R

        switch $R(class) {
            xCAMAC  {

                upvar 2 xCAMAC_Updateoptions options
                upvar 2 xCAMAC_Usage usage

                array set params [::cmdline::getKnownOptions Attributes $options $usage]
                
                foreach m [array names params] {
                    set rc 0
                    switch $m {
                        c  {if {$params(c) < 0 || $params(c) > 63} {set rc -1}}
                        n  {if {$params(n) <= 0 || $params(n) > 31} {set rc -1}}
                        a  {if {$params(a) < 0 || $params(a) > 31} {set rc -1}}
                        f  {
                           if {$params(f) < 0 || $params(f) >= 24} {set rc -1}
                           if {$params(f) >= 8 && $params(f) <= 15} {set rc -1}
                        }
                        q  {if {$params(q) != 0 && $params(q) != 1} {set rc -1}}
                        w  {if {$params(w) != 16 && $params(w) != 24} {set rc -1}}
                        p  {if {$params(p) != "rw" && $params(p) != "ro" && $params(p) != "wo"} {set rc -1}}
                        l  {if {$params(l) < 0 || $params(l) > 24} {set rc -1}}
                        b  {if {$params(b) < 0 || $params(b) > 24} {set rc -1}}
                        z  {if {$params(z) != "d" && $params(z) != "x" && $params(z) != "b"} {set rc -1}}
                        i  {}
                        default {set rc -1}
                    }
                    if {$params(p) == "wo" && $params(f) <= 7} {set rc -1}
                    if {$params(p) == "ro" && $params(f) > 7} {set rc -1}
                    if {$params(w) == 16 && $params(l) > 15} {set rc -1}
                    if {$params(w) == 16 && $params(b) > 15} {set rc -1}

                    if {$rc == 0 && $params($m) != 0xa55a} {set R($m) $params($m)}
                }
            }
            cCAMAC  {

                upvar 2 cCAMAC_Updateoptions options
                upvar 2 cCAMAC_Usage usage

                array set params [::cmdline::getKnownOptions Attributes $options $usage]
                
                foreach m [array names params] {
                    set rc 0
                    switch $m {
                        c  {if {$params(c) < 0 || $params(c) > 63} {set rc -1}}
                        n  {if {$params(n) <= 0 || $params(n) > 31} {set rc -1}}
                        a  {if {$params(a) < 0 || $params(a) > 31} {set rc -1}}
                        f  {
                           if {$params(f) <= 7 || $params(f) > 31} {set rc -1}
                           if {$params(f) >= 16 && $params(f) <= 23} {set rc -1}
                        }
                        q  {if {$params(q) != 0 && $params(q) != 1} {set rc -1}}
                        default {set rc -1}
                    }

                    if {$rc == 0 && $params($m) != 0xa55a} {set R($m) $params($m)}
                }
            }
            default {
                return "0x10004 Invalid request"
            }
        }
 
        return "0 OK"
    }

    proc do_WriteRegister {Register Value} {

        global MIDAS_CAMACAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames
        upvar 2 HasBeenConfigured HasBeenConfigured
        upvar 2 Debug Debug
        upvar 2 Address Address

        set Index $RegisterNames($Register)
        upvar 2 R$Index R

        switch $R(class) {
            xCAMAC  {
                if {$R(p) == "ro"} {
                    return "0x2004 Register is not writeable"
                }

                if {($R(b) != 0) || ($R(l) != 0) } then {    ;#   need to get the current value
                    set vl [do_ReadRegister $Register]
                    if {[first $vl] != 0} {return $vl}
                    set data [second $vl]

#    data now contains data last written or data read back
#    it is raw - not masked or shifted
#    mask only those bits not to be changed by this write
#    then mask out those bits to be changed by this write
#    and OR in this write data

                    set data [expr $data  & ~(~(~0 << $R(l)) << $R(b))]
                    set vl [expr (($Value & ~(~0 << $R(l))) << $R(b)) | $data]

                } else {set vl $Value}

                if {$R(p) == "rw"} {set f [expr ($R(f) & 15) + 16]} else {set f $R(f)}

                if {$MIDAS_CAMACAccessServer_Verbose} \
                    {puts "CAM request c=$R(c) n=$R(n) a=$R(a) f=$f w=$R(w) p=$R(p) l=$R(l) b=$R(b) data=$vl"}

                set Address "-c $R(c) -n $R(n) -a $R(a) -f $f"

                set rc [wtCAMAC -c $R(c) -n $R(n) -a $R(a) -f $f -d $vl]

                if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM reply $rc"}

                set R(LastWrite) $Value
            }
            cCAMAC  {
                if {$R(p) == "rw"} {set f [expr ($R(f) & 15) + 16]} else {set f $R(f)}

                if {$MIDAS_CAMACAccessServer_Verbose} \
                    {puts "CAM request c=$R(c) n=$R(n) a=$R(a) f=$f"}

                set Address "-c $R(c) -n $R(n) -a $R(a) -f $f"

                set rc [wtCAMAC -c $R(c) -n $R(n) -a $R(a) -f $f]

                if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM reply $rc"}
            }
            camAddr {
                set Address "$Value"
            }
            camExec  {
               if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM request $Address"}

                set rc [wtCAMAC -c $R(c) -n $R(n) -a $R(a) -f $R(f)]

                if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM reply $rc"}
            }
            camState {
                set HasBeenConfigured $Value
            }
            camDebug {
                set Debug $Value
            }
            default {
                return "0x10004 Invalid request"
            }
        }
 
        return "0 OK"
    }

    proc do_ReadRegister {Register} {

        global MIDAS_CAMACAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames
        if {![info exists RegisterNames]} {upvar 3 RegisterNames RegisterNames}
        upvar 2 HasBeenConfigured HasBeenConfigured
        if {![info exists HasBeenConfigured]} {upvar 3 HasBeenConfigured HasBeenConfigured}
        upvar 2 Debug Debug
        if {![info exists Debug]} {upvar 3 Debug Debug}
        upvar 2 Address Address
        if {![info exists Address]} {upvar 3 Address Address}
        upvar 2 Data Data
        if {![info exists Data]} {upvar 3 Data Data}
        upvar 2 Status Status
        if {![info exists Status]} {upvar 3 Status Status}

        set Index $RegisterNames($Register)
        upvar 2 R$Index R
        if {![info exists R]} {upvar 3 R$Index R}

        switch $R(class) {
            xCAMAC  {
                if {$R(p) == "wo"} then {
                    set Value $R(LastWrite)
                } else {
                    if {$MIDAS_CAMACAccessServer_Verbose} \
                        {puts "CAM request c=$R(c) n=$R(n) a=$R(a) f=$R(f) w=$R(w) p=$R(p) l=$R(l) b=$R(b)"}

                    set Address "-c $R(c) -n $R(n) -a $R(a) -f $R(f)"

                    set rc [rdCAMAC -c $R(c) -n $R(n) -a $R(a) -f $R(f)]

                    if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM reply $rc"}

                    set Value [second $rc]; set R(LastRead) $Value

                    if {($R(b) != 0) || ($R(l) != 0)} then {set Value [expr ($Value >> $R(b)) & ~(~0 << $R(l))]}

                    switch $R(z) {
                        x   {set Value [format "0x%0*x" [expr $R(w)/4] $Value]}
                        d   {set Value [format "%d" $Value]}
                        b   {}
                    }

                    if {$R(q) == 1} {
                        set QX [third $rc]
                        append Value " [format "%%%i%i" [expr ($QX>>1)&1] [expr $QX&1]]"
                    }
                }
            }
            cCAMAC  {
                if {$MIDAS_CAMACAccessServer_Verbose} \
                    {puts "CAM request c=$R(c) n=$R(n) a=$R(a) f=$R(f)"}

                set Address "-c $R(c) -n $R(n) -a $R(a) -f $R(f)"

                set rc [rdCAMAC -c $R(c) -n $R(n) -a $R(a) -f $R(f)]

                if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM reply $rc"}

                if {$R(q) == 1} {
                    set QX [third $rc]
                    set Value " [format "%%%i%i" [expr ($QX>>1)&1] [expr $QX&1]]
                } else {set Value ""}

            }
            camExec  {
                if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM request $Address"}

                set rc [rdCAMAC $Address]

                if {$MIDAS_CAMACAccessServer_Verbose}  {puts "CAM reply $rc"}

                set Value [second $rc]
            }
            camState {
                set Value $HasBeenConfigured
            }
            camStatus {
                set Value $Status
            }
            camDebug {
                set Value $Debug
            }
            camAddr {
                set Value "$Address"
            }
            camData {
                set Value $Data
            }
            default {
                return "0x10004 Invalid request"
            }
        }
        return "0 $Value"
    }

    proc InterfaceStatus {} {

        set Response [stCAMAC]    ;#  get interface specific response 

        return [rpcvar::rpcvar string "$Response"]
    }


    # We have to publish the public methods...

     SOAP::export AccessServerVerbose
     SOAP::export DefineRegister
     SOAP::export InquireRegisters
     SOAP::export ReadRegisterAttributes
     SOAP::export ReadRegistersAttributes
     SOAP::export DeleteRegister
     SOAP::export DeleteRegisters
     SOAP::export WriteRegisterAttributes
     SOAP::export WriteRegistersAttributes
     SOAP::export WriteRegister
     SOAP::export WriteRegisters
     SOAP::export InitialiseRegister
     SOAP::export InitialiseRegisters
     SOAP::export ReadRegister
     SOAP::export ReadRegisters
     SOAP::export InterfaceStatus

     
}

# register this service with tclhttpd
puts [SOAP::Domain::register  -prefix  /CAMACAccessServer -namespace urn:CAMACAccessServer -uri urn:CAMACAccessServer]


proc ::SOAP::Domain::MIDAS_SOAPDomainHandler_/CAMACAccessServer {optsname sock args} {   \
    ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DefaultAccessServer $optsname $sock $args}

