#    MIDAS Access Server  for  Spectrum 

global env
source [file join $env(MIDASBASE) TclHttpd tcl MIDASDomainHandler.tcl]       ;#  load default support procedures

source [file join $env(MIDASBASE) packages Spectrum pkgIndex.tcl]
package require Spectrum

global MIDAS_SpectrumServer_Verbose
set MIDAS_SpectrumServer_Verbose 0


# Use namespaces to isolate your methods
namespace eval urn:SpectrumServer {
  catch {namespace import ::Spectrum::New1D}
  catch {namespace import ::Spectrum::New2D}
  catch {namespace import ::Spectrum::Delete}
  catch {namespace import ::Spectrum::Zero}
  catch {namespace import ::Spectrum::Details}
  catch {namespace import ::Spectrum::Names}
  catch {namespace import ::Spectrum::Read1D}
  catch {namespace import ::Spectrum::Read2D}
  catch {namespace import ::Spectrum::Title}
  catch {namespace import ::Spectrum::Label}
  catch {namespace import ::Spectrum::Info}
  catch {namespace import ::Spectrum::Calib}

  proc SpectrumServerVerbose {option} {
      global MIDAS_SpectrumServer_Verbose
      global MIDAS_SOAPDomainHandler_Verbose
      set MIDAS_SpectrumServer_Verbose $option
      set MIDAS_SOAPDomainHandler_Verbose $option
      return [rpcvar::rpcvar string "0 OK"]
  }

    proc SpecNew1D {n b r} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecNew1D with parameters $n $b $r"}
        set z [catch {New1D $n $b $r} m]
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecNew2D {n b0 r0 b1 r1} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecNew2D with parameters $n $b0 $r0 $b1 $r1"}
        set z [catch {New2D $n $b0 $r0 $b1 $r1} m]
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecDelete {n} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecDelete with parameter $n"}
        set z [catch {Delete $n} m]
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecZero {n} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecZero with parameter $n"}
        set z [catch {Zero $n} m]
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecDetails {n} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecDetails with parameter $n"}
        set z [catch {Details $n} m]
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecNames {args} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecNames with parameter $args"}
        set z [catch {Names $args} m]
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecRead1D {n args} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecRead1D with parameters $n $args"}
        if {[llength $args] < 2} {
           set z [catch {Details $n} d]
           if {[llength $args] == 1} {
             set b [lindex $args 0]
             set r [expr [lindex $d 1] + [lindex $d 2] - $b]
           } else {
             set b [lindex $d 1]
             set r [lindex $d 2]
           }
         } else {
           set b [lindex $args 0]
           set r [lindex $args 1]
         }
        if {$MIDAS_SpectrumServer_Verbose} {puts "calling method Spectrum::Read1D with parameters $n $b $r"}
        set z [catch {Read1D $n $b $r} d]
        if {!$z} {set z [catch {::base64::encode $d} m]
        } else {set m $d}

        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecRead2D {n args} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecRead2D with parameters $n $args"}
        if {[llength $args] < 4} {
           set z [catch {Details $n} d]
           if {[llength $args] == 0} {
             set b1 [lindex $d 1]
             set r1 [lindex $d 2]
             set b2 [lindex $d 3]
             set r2 [lindex $d 4]
           }
           if {[llength $args] == 1} {
             set b1 [lindex $args 0]
             set r1 [expr [lindex $d 1] + [lindex $d 2] - $b1]
             set b2 [lindex $d 3]
             set r2 [lindex $d 4]
           }
           if {[llength $args] == 2} {
             set b1 [lindex $args 0]
             set r1 [lindex $args 1]
             set b2 [lindex $d 3]
             set r2 [lindex $d 4]
           }
           if {[llength $args] == 3} {
             set b1 [lindex $args 0]
             set r1 [lindex $args 1]
             set b2 [lindex $args 2]
             set r2 [expr [lindex $d 3] + [lindex $d 4] - $b2]
           }
         } else {
           set b1 [lindex $args 0]
           set r1 [lindex $args 1]
           set b2 [lindex $args 2]
           set r2 [lindex $args 3]
         }
        if {$MIDAS_SpectrumServer_Verbose} {puts "calling method Spectrum::Read1D with parameters $n $b1 $r1 $b2 $r2"}
        set z [catch {Read2D $n $b1 $r1 $b2 $r2} d]
        set z [catch {::base64::encode $d} m]

        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecTitle {s args} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecTitle with parameters $s $args"}
        if {[llength $args] == 0} {
           set z [catch {Title $s} m]
        } else {
           set t [lindex $args 0]
           set z [catch {Title $s $t} m]
        }
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecLabel {s args} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecLabel with parameters $s $args"}
        if {[llength $args] == 0} {
           set z [catch {Label $s} m]
        } else {
           set l [lindex $args 0]
           set z [catch {Label $s $l} m]
        }
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecInfo {s n args} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecInfo with parameters $s $n $args"}
        if {[llength $args] == 0} {
           set z [catch {Info $s $n} m]
        } else {
           set i [lindex $args 0]
           set z [catch {Info $s $n $i} m]
        }
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }

    proc SpecCalib {s d args} {
        global   MIDAS_SpectrumServer_Verbose
        if {$MIDAS_SpectrumServer_Verbose} {puts "called method SpecCalib with parameters $s $d $args"}
        if {[llength $args] == 0} {
           set z [catch {Calib $s $d} m]
        } else {
           set c [lindex $args 0]
           set z [catch {Calib $s $d $c} m]
        }
        set status "$m"
        return [rpcvar::rpcvar string $status]
    }


    # We have to publish the public methods...

     SOAP::export SpectrumServerVerbose
     SOAP::export SpecNew1D
     SOAP::export SpecNew2D
     SOAP::export SpecDelete
     SOAP::export SpecZero
     SOAP::export SpecDetails
     SOAP::export SpecNames
     SOAP::export SpecRead1D
     SOAP::export SpecRead2D
     SOAP::export SpecTitle
     SOAP::export SpecLabel
     SOAP::export SpecInfo
     SOAP::export SpecCalib
}

# register this service with tclhttpd
puts [SOAP::Domain::register  -prefix  /SpectrumServer -namespace urn:SpectrumServer -uri urn:SpectrumServer]

proc ::SOAP::Domain::MIDAS_SOAPDomainHandler_/SpectrumServer {optsname sock args} {   \
    ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DefaultAccessServer $optsname $sock $args}
