
#    MIDAS Access Server  for  VME 

global env
source [file join $env(MIDASBASE) TclHttpd tcl MIDASDomainHandler.tcl]       ;#  load default support procedures

#     Note that the method of VME access (e.g. via the CAEN V1718 USB2 interface) is defined in the server startup file

global MIDAS_VMEAccessServer_Verbose
set MIDAS_VMEAccessServer_Verbose 0

global MIDAS_SOAPDomainHandler_Verbose
set MIDAS_SOAPDomainHandler_Verbose 0

global MIDAS_RegisterAccessServer_Verbose
set MIDAS_RegisterAccessServer_Verbose 0


# Use namespaces to isolate your methods
namespace eval urn:VMEAccessServer {

global env
source [file join $env(MIDASBASE) TclHttpd tcl RegisterAccessServer.tcl]       ;#  load common procedures

#    set up the Instrinsic registers

    set HasBeenConfigured 0
    set Address ""
    set Data 0
    set Status 0
    set Debug 0

    set RegisterIndex 1
    
    set RegisterNames(HasBeenConfigured) $RegisterIndex
    set R[set RegisterIndex](class) vmeState
    incr RegisterIndex

    set RegisterNames(VME.Address) $RegisterIndex
    set R[set RegisterIndex](class) vmeAddr
    incr RegisterIndex

    set RegisterNames(VME.Execute) $RegisterIndex
    set R[set RegisterIndex](class) vmeExec
    incr RegisterIndex

    set RegisterNames(VME.Data) $RegisterIndex
    set R[set RegisterIndex](class) vmeData
    incr RegisterIndex

    set RegisterNames(VME.Status) $RegisterIndex
    set R[set RegisterIndex](class) vmeStatus
    incr RegisterIndex

    set RegisterNames(VME.Debug) $RegisterIndex
    set R[set RegisterIndex](class) vmeDebug
    incr RegisterIndex


#    xVME register attributes

   set xVME_options {
         {a.arg 0    "VME address base"}
         {o.arg 0    "VME address offset"}
         {w.arg 32   "address width"}
         {d.arg 32   "data width"}
         {u.arg "d"  "use"}
         {m.arg 0x09  "mode or VME address modifier"}
         {p.arg "rw" "access permission"}
         {b.arg 0    "field width (bits)"}
         {l.arg 0    "field length (bits)"}
         {z.arg "x"  "data format returned"}
         {s.arg "0"  "return status"}
         {i.arg 0xa55a  "initial value"}
   }

   set xVME_Updateoptions {
         {a.arg 0xa55a    "VME address base"}
         {o.arg 0xa55a    "VME address offset"}
         {w.arg 0xa55a    "address width"}
         {d.arg 0xa55a    "data width"}
         {u.arg 0xa55a    "use"}
         {m.arg 0xa55a    "mode or VME address modifier"}
         {p.arg 0xa55a    "access permission"}
         {b.arg 0xa55a    "field width (bits)"}
         {l.arg 0xa55a    "field length (bits)"}
         {z.arg 0xa55a    "data format returned"}
         {s.arg 0xa55a    "return status"}
         {i.arg 0xa55a    "initial value"}
    }

    set xVME_Usage ": xVME  -a ? -o ? -w 16|24|32 -d 8|16|32 -u d|p -m n|s|? -p ro|wo|rw -b ? -l ? -z w|b|d -s 0|1 -i ?"


#    bVME register attributes

   set bVME_options {
         {a.arg 0    "VME address base"}
         {o.arg 0    "VME address offset"}
         {w.arg 32   "address width"}
         {d.arg 32   "data width"}
         {u.arg "d"  "use"}
         {m.arg 0x09  "mode or VME address modifier"}
         {p.arg "rw" "access permission"}
         {z.arg "x"  "data format returned"}
         {l.arg "1"  "length in data items"}
         {i.arg 0xa55a  "initial value"}
   }

   set bVME_Updateoptions {
         {a.arg 0xa55a    "VME address base"}
         {o.arg 0xa55a    "VME address offset"}
         {w.arg 0xa55a    "address width"}
         {d.arg 0xa55a    "data width"}
         {u.arg 0xa55a    "use"}
         {m.arg 0xa55a    "mode or VME address modifier"}
         {p.arg 0xa55a    "access permission"}
         {z.arg 0xa55a    "data format returned"}
         {l.arg 0xa55a    "length in data items"}
         {i.arg 0xa55a    "initial value"}
    }

    set bVME_Usage ": xVME  -a ? -o ? -w 16|24|32 -d 8|16|32 -u d|p -m n|s|? -p ro|wo|rw -z w|b|d -l ? -i ?"


    proc AccessServerVerbose {option} {
        global MIDAS_VMEAccessServer_Verbose
        global MIDAS_RegisterAccessServer_Verbose
        global MIDAS_SOAPDomainHandler_Verbose
        set MIDAS_VMEAccessServer_Verbose $option
        set MIDAS_RegisterAccessServer_Verbose $option
        set MIDAS_SOAPDomainHandler_Verbose $option
        return [rpcvar::rpcvar string "0 OK"]
    }

    proc DefineRegister {Register Class Attributes} {
        global MIDAS_VMEAccessServer_Verbose
        upvar RegisterIndex RegisterIndex
        upvar RegisterNames RegisterNames

        if {$MIDAS_VMEAccessServer_Verbose} \
             {puts "called method VMEAccessServer::DefineRegister with parameters $Register $Class $Attributes"}
             
#    Check Register name is not already in use

        if {[info exists RegisterNames($Register)]} then {
             return [rpcvar::rpcvar string "0x10006 Register name already in use"]
        }

#    Check Class is valid

        switch $Class {
            xVME -
            bVME  {}

            default {
               return [rpcvar::rpcvar string "0x10007 Class name is invalid or not supported"]
            }
        }

#    create register

        switch $Class {
            xVME  {

                set RegisterNames($Register) $RegisterIndex

                upvar R$RegisterIndex R
                set R(class) $Class

                upvar xVME_options options

                foreach item $options {
                    set opt [first [split [first $item] .]]
                    set R($opt) [second $item]
                }

                set R(LastRead)  0x0
                set R(LastWrite) 0x0

                incr RegisterIndex
            }
            bVME  {

                set RegisterNames($Register) $RegisterIndex

                upvar R$RegisterIndex R
                set R(class) $Class

                upvar bVME_options options

                foreach item $options {
                    set opt [first [split [first $item] .]]
                    set R($opt) [second $item]
                }
                
                incr RegisterIndex
            }
            default {
                return [rpcvar::rpcvar string "0x10004 Invalid request"]
            }
        }

#    apply attributes

        set response [do_WriteRegisterAttributes $Register $Attributes]

        return [rpcvar::rpcvar string $response]
    }


    proc do_ReadRegisterAttributes {Register} {

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames

        set Index $RegisterNames($Register)
        upvar 2 R$Index R

        switch $R(class) {
            xVME {
                 set attributes "-a$R(a) -o$R(o) -w$R(w) -d$R(d) -m$R(m) -p$R(p) -l$R(l) -b$R(b) -z$R(z) -s$R(s)"
                 if {$R(i) != 0xa55a} {append attributes " -i$R(i)"}
            }
            bVME {
                 set attributes "-a$R(a) -o$R(o) -w$R(w) -d$R(d) -m$R(m) -p$R(p) -z$R(z) -l$R(l)"
                 if {$R(i) != 0xa55a} {append attributes " -i$R(i)"}
            }
            default {
                return "0x10004 Invalid request"
            }
        }

        return "0 $attributes"
    }

    proc scan-num s {    ;# remove after common.tcl updated
       if {[string first "0x" $s]==0} then {
         set z [scan [string range $s 2 end] %x zzz]
       } else {
         set z [scan $s %f zzz]
       }
       if {$z==1} then {return $zzz} else {return "???"}
    }

    proc do_WriteRegisterAttributes {Register Attributes} {

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames

        set Index $RegisterNames($Register)
        upvar 2 R$Index R

        switch $R(class) {
            xVME  {

                upvar 2 xVME_Updateoptions options
                upvar 2 xVME_Usage usage

                array set params [::cmdline::getKnownOptions Attributes $options $usage]
                
                foreach m [array names params] {
                    set rc 0
                    switch $m {
                        a  {}
                        o  {}
                        w  {if {$params(w) != 16 && $params(w) != 24 && $params(w) != 32} {set rc -1}}
                        d  {if {$params(w) != 8 && $params(w) != 16 && $params(w) != 32} {set rc -1}}
                        u  {if {$params(u) != "d" && $params(u) != "p"} {set rc -1}}
                        m  {
                           if {[scan-num $params(m)] == "???"} then {
                               if {$params(m) != "n" && $params(m) != "s"} {set rc -1}
                           }
                        }
                        p  {if {$params(p) != "rw" && $params(p) != "ro" && $params(p) != "wo"} {set rc -1}}
                        l  {if {$params(l) < 0 || $params(l) > 32} {set rc -1}}
                        b  {if {$params(b) < 0 || $params(b) > 32} {set rc -1}}
                        z  {if {$params(z) != "d" && $params(z) != "x" && $params(z) != "b"} {set rc -1}}
                        i  {}
                        s  {if {$params(s) != 0 && $params(s) != 1} {set rc -1}}
                        default {set rc -1}
                    }
                    if {$rc == 0 && $params($m) != 0xa55a} {set R($m) $params($m)}
                }
                if {[scan-num $params(m)] == "???"} {
                    switch [set params(u)][set params(m)][set params(w)] {
                        dn16 {set params(m) 0x29}
                        ds16 {set params(m) 0x2d}
                        pn16 {set params(m) 0x29}
                        ps16 {set params(m) 0x2d}
                        dn24 {set params(m) 0x39}
                        ds24 {set params(m) 0x3d}
                        pn24 {set params(m) 0x3a}
                        ps24 {set params(m) 0x3e}
                        dn32 {set params(m) 0x09}
                        ds32 {set params(m) 0x0d}
                        pn32 {set params(m) 0x0a}
                        ps32 {set params(m) 0x0e}
                        default {set params(m) 0x09}
                    }
                } else {
                    switch [format 0x%02x $params(m)] {
                        0x29 -
                        0x2d -
                        0x39 -
                        0x3d -
                        0x3a -
                        0x3e -
                        0x09 -
                        0x0d -
                        0x0a -
                        0x0e  {set params(m) [format 0x%02x $params(m)]}
                        default  {set params(m) 0x09}
                    }
                }
            }
            bVME  {

                upvar 2 cVME_Updateoptions options
                upvar 2 cVME_Usage usage

                array set params [::cmdline::getKnownOptions Attributes $options $usage]
                
                foreach m [array names params] {
                    set rc 0
                    switch $m {
                        a  {}
                        o  {}
                        w  {if {$params(w) != 16 && $params(w) != 24 && $params(w) != 32} {set rc -1}}
                        d  {if {$params(w) != 8 && $params(w) != 16 && $params(w) != 32} {set rc -1}}
                        u  {if {$params(u) != "d" && $params(u) != "p"} {set rc -1}}
                        m  {
                           if {[scan-num $params(m)] == "???"} then {
                               if {$params(m) != "n" && $params(m) != "s"} {set rc -1}
                           }
                        }
                        p  {if {$params(p) != "rw" && $params(p) != "ro" && $params(p) != "wo"} {set rc -1}}
                        z  {if {$params(z) != "d" && $params(z) != "x" && $params(z) != "b"} {set rc -1}}
                        l  {if {$params(l) < 1 || $params(l) > 1024} {set rc -1}}
                        i  {}
                        default {set rc -1}
                    }
                    if {$rc == 0 && $params($m) != 0xa55a} {set R($m) $params($m)}
                }
                if {[scan-num $params(m)] == "???"} {
                    switch [set params(u)][set params(m)][set params(w)] {
                        dn16 {set params(m) 0x29}
                        ds16 {set params(m) 0x2d}
                        pn16 {set params(m) 0x29}
                        ps16 {set params(m) 0x2d}
                        dn24 {set params(m) 0x39}
                        ds24 {set params(m) 0x3d}
                        pn24 {set params(m) 0x3a}
                        ps24 {set params(m) 0x3e}
                        dn32 {set params(m) 0x09}
                        ds32 {set params(m) 0x0d}
                        pn32 {set params(m) 0x0a}
                        ps32 {set params(m) 0x0e}
                        default {set params(m) 0x09}
                    }
                } else {
                    switch [format 0x%02x $params(m)] {
                        0x29 -
                        0x2d -
                        0x39 -
                        0x3d -
                        0x3a -
                        0x3e -
                        0x09 -
                        0x0d -
                        0x0a -
                        0x0e  {set params(m) [format 0x%02x $params(m)]}
                        default  {set params(m) 0x09}
                    }
                }
            }
            default {
                return "0x10004 Invalid request"
            }
        }
 
        return "0 OK"
    }

    proc do_WriteRegister {Register Value} {

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames
        upvar 2 HasBeenConfigured HasBeenConfigured
        upvar 2 Debug Debug
        upvar 2 Address Address

        set Index $RegisterNames($Register)
        upvar 2 R$Index R

        switch $R(class) {
            xVME  {
                if {$R(p) == "ro"} {
                    return "0x2004 Register is not writeable"
                }

                if {($R(b) != 0) || ($R(l) != 0) } then {    ;#   need to get the current value
                    set vl [do_ReadRegister $Register]
                    if {[first $vl] != 0} {return $vl}
                    set data [second $vl]

#    data now contains data last written or data read back
#    it is raw - not masked or shifted
#    mask only those bits not to be changed by this write
#    then mask out those bits to be changed by this write
#    and OR in this write data

                    set data [expr $data  & ~(~(~0 << $R(l)) << $R(b))]
                    set vl [expr (($Value & ~(~0 << $R(l))) << $R(b)) | $data]

                } else {set vl $Value}

                if {$MIDAS_VMEAccessServer_Verbose} \
                    {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) m=$R(m) data=$vl"}

                set Addr [expr $R(a) + $R(o)]
                set DataWidth $R(d)
                set AddrMod $R(m)

                set rc [wtVME -a $Addr -w $DataWidth -m $AddrMod -d $vl]

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                set R(LastWrite) $Value
            }
            bVME  {
                if {$R(p) == "ro"} {
                    return "0x2004 Register is not writeable"
                }
                if {$MIDAS_VMEAccessServer_Verbose} \
                    {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) m=$R(m) data=$vl"}

                set Addr [expr $R(a) + $R(o)]
                set DataWidth $R(d)
                set AddrMod $R(m)

                set rc [wtblkVME -a $Addr -w $DataWidth -m $AddrMod -d $vl]

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

            }
            vmeAddr {
                set Address "$Value"
            }
            vmeExec  {
               if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME request $Address"}

                set rc [wtVME -a $Addr -w $DataWidth -m $AddrMod -d $Value]

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}
            }
            vmeState {
                set HasBeenConfigured $Value
            }
            vmeDebug {
                set Debug $Value
            }
            default {
                return "0x10004 Invalid request"
            }
        }
 
        return "0 OK"
    }

    proc do_ReadRegister {Register} {

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames
        if {![info exists RegisterNames]} {upvar 3 RegisterNames RegisterNames}
        upvar 2 HasBeenConfigured HasBeenConfigured
        if {![info exists HasBeenConfigured]} {upvar 3 HasBeenConfigured HasBeenConfigured}
        upvar 2 Debug Debug
        if {![info exists Debug]} {upvar 3 Debug Debug}
        upvar 2 Address Address
        if {![info exists Address]} {upvar 3 Address Address}
        upvar 2 Data Data
        if {![info exists Data]} {upvar 3 Data Data}
        upvar 2 Status Status
        if {![info exists Status]} {upvar 3 Status Status}

        set Index $RegisterNames($Register)
        upvar 2 R$Index R
        if {![info exists R]} {upvar 3 R$Index R}

        switch $R(class) {
            xVME  {
                if {$R(p) == "wo"} then {
                    set Value $R(LastWrite)
                } else {
                    if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) u=$R(u) m=$R(m)"}

                    set Addr [expr $R(a) + $R(o)]
                    set DataWidth $R(d)
                    set AddrMod $R(m)

                    set rc [rdVME -a $Addr -w $DataWidth -m $AddrMod]

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                    set Value [second $rc]; set R(LastRead) $Value

                    if {($R(b) != 0) || ($R(l) != 0)} then {set Value [expr ($Value >> $R(b)) & ~(~0 << $R(l))]}

                    switch $R(z) {
                        x   {set Value [format "0x%0*x" [expr $R(w)/4] $Value]}
                        d   {set Value [format "%d" $Value]}
                        b   {}
                    }
                }
            }
            bVME  {
                if {$R(p) == "wo"} then {
                    return "0x2004 Register is not readable"
                } else {
                    if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) m=$R(m) l=$R(l)"}

                    set Addr [expr $R(a) + $R(o)]
                    set DataWidth $R(d)
                    set AddrMod $R(m)

                    set rc [rdblkVME -a $Addr -w $DataWidth -m $AddrMod -l $R(l)]

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                    set Value [second $rc]

#                    switch $R(z) {
#                        x   {set Value [format "0x%0*x" [expr $R(w)/4] $Value]}
#                        d   {set Value [format "%d" $Value]}
#                        b   {}
#                    }
                }
            }
            vmeExec  {
                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME request $Address"}

                set rc [rdVME $Address]

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                set Value [second $rc]
            }
            vmeState {
                set Value $HasBeenConfigured
            }
            vmeStatus {
                set Value $Status
            }
            vmeDebug {
                set Value $Debug
            }
            vmeAddr {
                set Value "$Address"
            }
            vmeData {
                set Value $Data
            }
            default {
                return "0x10004 Invalid request"
            }
        }
        return "0 $Value"
    }

    proc InterfaceStatus {} {

        set Response [stVME]    ;#  get interface specific response 

        return [rpcvar::rpcvar string "$Response"]
    }


    # We have to publish the public methods...

     SOAP::export AccessServerVerbose
     SOAP::export DefineRegister
     SOAP::export InquireRegisters
     SOAP::export ReadRegisterAttributes
     SOAP::export ReadRegistersAttributes
     SOAP::export DeleteRegister
     SOAP::export DeleteRegisters
     SOAP::export WriteRegisterAttributes
     SOAP::export WriteRegistersAttributes
     SOAP::export WriteRegister
     SOAP::export WriteRegisters
     SOAP::export InitialiseRegister
     SOAP::export InitialiseRegisters
     SOAP::export ReadRegister
     SOAP::export ReadRegisters
     SOAP::export InterfaceStatus

     
}

# register this service with tclhttpd
puts [SOAP::Domain::register  -prefix  /VMEAccessServer -namespace urn:VMEAccessServer -uri urn:VMEAccessServer]


proc ::SOAP::Domain::MIDAS_SOAPDomainHandler_/VMEAccessServer {optsname sock args} {   \
    ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DefaultAccessServer $optsname $sock $args}

